/**
 * Class Complex - Implements complex numbers, and manipulation functions.
 * Many of these methods have been designed to avoid unnnessecary object
 * construction, so as to increase the speed of these operations. 
 * Results are either stored in parameters, or reflected in the actual object.
 *
 * @author Matt Johnston
 */

public class Complex {

private double real;
private double imag;


/**
 * Creates a new Complex, with real and imaginary parts both zero.
 */
public Complex() {

	real = 0;
	imag = 0;

}

/**
 * Creates a new Complex, with specified components.
 * @param re The real component of the new complex number.
 * @param im The imaginary component of the new complex number.
 */
public Complex(double re, double im) {

	real = re;
	imag = im;

}

/**
 * Creates a new Complex, with the same value as an existing
 * complex number.
 * @param x An existing complex number to use as the basis for
 * the new complex number. 
 */
public Complex(Complex x) {

	real = x.real;
	imag = x.imag;

}

/**
 * Sets the real part of a complex number to a specified value.
 * @param re The value to set the real part of the complex number.
 */
public void setRe(double re) {

	real = re;

}

/**
 * Sets the imaginary part of a complex number to a specified value
 * @param im The value to set the imaginary part of the complex number.
 */
public void setIm(double im) {

	imag = im;

}

/**
 * Assigns the value of another Complex number to this Complex number.
 * This is useful for reusing Complex objects to increase efficiency.
 * @param x The complex number to assign to this object.
 */
public void setComplex(Complex x) {

	real = x.real;
	imag = x.imag;

}

/**
 * Sets both real and imaginary parts of the complex number to zero.
 * This is useful for efficiency, just reset a current Complex rather
 * than creating a new object, which is much slower.
 */
public void setZero() {

	real = 0;
	imag = 0;

}

/**
 * Retrieves the real part of the complex number.
 * @return The real part of the complex number.
 */
public double Re() {

	return real;

}

/**
 * Retrieves the imaginary part of the complex number.
 * @return The imaginary part of the complex number.
 */
public double Im() {

	return imag;

}

/**
 * Adds another complex number to the current complex number.
 * The result of this function is that this complex number
 * contains the sum of the old complex number, and the parameter.
 * @param x The complex number to add.
 */
public void add(Complex x) {

	real = real + x.real;
	imag = imag + x.imag;

}

/**
 * Adds a real value to the current complex number. This is similar to
 * the add method, except the number being added doesn't have to be
 * a Complex number. The result of the operation is stored as the value
 * of this object.
 * @param x The real number to add.
 */
public void addReal(double x) {

	real = real + x;

}

/**
 * Subtracts a complex number from this complex number. The result is 
 * stored as the value of this object.
 * @param x A complex number to subtract.
 */
public void subtract(Complex x) {

	real = real - x.real;
	imag = imag - x.imag;

}

/**
 * Multiplies this complex number by a real number. The result is 
 * returned as a new Complex.
 * @param k The real number to multiply this by.
 * @return A new Complex number representing the result.
 */
public Complex scalarmul(double k) {

	return new Complex(real*k, imag*k);

}

/**
 * Multiplies this complex number by a real number. The result is 
 * stored as the value of the object.
 * @param k The real number to multiply this by.
 */
public void scalarmul(double k, Complex output) {

	output.real = real * k;
	output.imag = imag * k;

}

/**
 * Multiplies this complex number by a complex number. The result is returned
 * as a new complex number.
 * @param x The complex number to multiply this by.
 * @return A new complex number which is the multiple of this and x.
 */
public Complex multiply(Complex x) {

	return new Complex(real*x.real - imag*x.imag, real*x.imag + imag*x.real);

}

/**
 * Multiplies this complex number by a complex number. The result is stored
 * in the parameter "output".
 * @param x The complex number to multiply by.
 * @param output An initialised Complex object to store the result in.
 */
public void multiply(Complex x, Complex output) {

	double re = real*x.real - imag*x.imag;
	double im = real*x.imag + imag*x.real;

	output.real = re;
	output.imag = im;

}

/**
 * Divides this complex number by a complex number.
 * @param x The complex number to divide this by.
 * @return A new complex number which is equal to this divided by x.
 * If a divide by zero error occurs, null is returned.
 */
public void divide(Complex x, Complex output) {

	double xmod = x.modulusSquare();

	if (xmod == 0) {
		output = null;
		return;
	}

	x.conjugate(output);
	this.multiply(output,output);

	output.scalarmul(1/xmod, output);

}

/**
 * Calculates the square of the modulus of this complex number. This is used
 * for testing for a zero divisor in the divide method.
 * @return The square of the modulus of this complex number.
 */
public double modulusSquare() {

	return real*real + imag*imag;

}

/**
 * Returns the conjugate of this complex number. The result is stored
 * in a parameter output.
 * @param output  An initialised Complex object to store the conjugate of
 * this.
 */
public void conjugate(Complex output) {

	output.real = real;
	output.imag = -imag;

}

/**
 * Raises this to the power of a specified integer.
 * @param pow The integer to raise this to. This must be non-negative.
 * @return A new complex number equal to this raised to the power of pow.
 * If pow is negative, null is returned.
 */
public Complex powerOf(int pow) {

	if (pow < 0) return null;

	Complex output = new Complex(1,0);

	for (int i = 1; i <= pow; i++)
		output = output.multiply(this);

	return output;
}	

/**
 * Returns a list of this complex number, equal to this complex number raised to sequential powers.
 * @param deg The extent of powers to return.
 * @return A list of complex numbers, equal to 
 * this^0, this^1, this^2, ... , this^deg
 */
public Complex[] powerList(int deg) {

	Complex[] output = new Complex[deg+1];

	output[0] = new Complex(1,0);

	for (int i = 1; i <= deg; i++)
		output[i] = output[i-1].multiply(this);

	return output;

}

}
